{#-
 # Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 # Copyright (C) 2021  UAVCAN Development Team  <uavcan.org>
 # This software is distributed under the terms of the MIT License.
-#}
{%- extends "base.j2" %}
{%- block object -%}
{% set composite_type = T %}{% include '_composite_type.j2' %}
{%- endblock -%}
