{#-
 # Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 # Copyright (C) 2021  UAVCAN Development Team  <uavcan.org>
 # This software is distributed under the terms of the MIT License.
-#}
{%- for field in composite_type.fields_except_padding %}
    {% if loop.first %}
    // +----------------------------------------------------------------------+
    // | FIELDS
    // +----------------------------------------------------------------------+
    {% endif -%}
    {{ field.doc | block_comment('cpp-doxygen', 4, 120) }}
    {{ field.data_type | declaration }} {{ field | id }};
{%- endfor -%}
