#
# Copyright (C) 2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Contains the Jinja templates to generate HTML documentation.
"""

__version__ = "1.0.0"
"""Version of the HTML docgen template set."""
